/****************************** Module Header *******************************
*
* Module Name: locate.e
*
* Copyright (c) Netlabs EPM Distribution Project 2002
*
* $Id: locate.e 3734 2019-12-07 13:43:57Z aschn $
*
* ===========================================================================
*
* This file is part of the Netlabs EPM Distribution package and is free
* software.  You can redistribute it and/or modify it under the terms of the
* GNU General Public License as published by the Free Software
* Foundation, in version 2 as it comes in the "COPYING" file of the
* Netlabs EPM Distribution.  This library is distributed in the hope that it
* will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
****************************************************************************/

; Undocumented:
; display -8 ==> messages go only to the msg line, not to the msg box.
; display 8  ==> reenables messages from a previous display -8
; The rest is documented in epmtech.inf.


; 'xcom l', 'xcom c' and repeatfind:
;
; o  'xcom l <delim><SearchString><delim><SearchOptions>' or
;    'xcom c <delim><SearchString><delim><ChangeString><delim><SearchOptions>'
;    can be specified without <delim><SearchOptions>. Then the internal
;    default options are used: +FAE.
;
; o  Doubled or exclusive options may be specified. The last option wins.
;
; o  'xcom l' and 'xcom c' store their command line in the same internal var.
;    It can be queried with getsearch and be set with setsearch. That var is
;    empty on EPM's start.
;
; o  repeatfind uses the previous search line of 'xcom l' or 'xcom c', but
;    always executes 'xcom l'. It can't be used for repeated change.


; Bugs in the original defs, now workarounded:
;
; o  Bugs of 'xcom l', 'xcom c' and repeatfind:
;    -  'xcom l' and 'xcom c' don't move the cursor, repeatfind does.
;    -  With option R (reversed search), it also finds the search string at
;       the cursor.
;    -  For search within a mark (option M), the search is restarted from
;       the mark begin when the end was reached.
;    -  For search within a char mark, the entire lines where the mark starts
;       and ends are searched, not only the marked area.
;       Note that a workaround was added for 'locate'. It does not fully work
;       with 'change'. A workaround for 'change all' doesn't seem to be
;       possible.
;
; o  Bugs of the original 'Locate' = 'l' and 'Change' = 'c' commands:
;    -  The bugs of 'xcom l', 'xcom c' and repeatfind, except that the
;       cursor is moved after 'l' or 'c'.
;    -  The cursor position after a 'c' was wrong: It was moved by the length
;       of the search string, not of the replace string.
;
; o  The above bugs are fixed/workarounded by the current 'Locate' = 'l'
;    and 'Change' = 'c' commands.
;
; o  Without pre- and postprocessing, 'xcom l', 'xcom c' and repeatfind
;    should only be used for forward find (locate) searches and not for
;    replace (change).

; ---------------------------------------------------------------------------

compile if not defined( SMALL)  -- If compiled separately
   EA_comment 'This defines macros for saerch operations.'

define
   INCLUDING_FILE = 'LOCATE.E'

   include 'stdconst.e'
const
   tryinclude 'mycnf.e'
const
 compile if not defined( NLS_LANGUAGE)
   NLS_LANGUAGE = 'ENGLISH'
 compile endif
   include NLS_LANGUAGE'.e'
compile endif

; ---------------------------------------------------------------------------
const
compile if not defined( LOCATE_CIRCLE_STYLE)
   --LOCATE_CIRCLE_STYLE = 1  -- changed by aschn
   LOCATE_CIRCLE_STYLE = 5         -- (1)     filled oval
compile endif
compile if not defined( LOCATE_CIRCLE_COLOR1)
   --LOCATE_CIRCLE_COLOR1 = 16777220  -- changed by aschn
   LOCATE_CIRCLE_COLOR1 = 16777231 -- (16777220) complementary
compile endif
compile if not defined( LOCATE_CIRCLE_COLOR2)
   -- for styles 2 and 4 only
   --LOCATE_CIRCLE_COLOR2 = 16777218  -- changed by aschn
   LOCATE_CIRCLE_COLOR2 = 16777216 -- (16777218) complementary
compile endif
compile if not defined( HIGHLIGHT_COLOR)
   HIGHLIGHT_COLOR = 14            --         This must be set to enable circle colors
compile endif

; ---------------------------------------------------------------------------
defmodify             -- This stops the modification dialog for grep output "files"  -- JBS
   if leftstr( .filename, 17) = '.Output from grep' then
      .modify = 0
      .autosave = 0
   endif

; ---------------------------------------------------------------------------
; Preprocessor 1 for Locate and Change. It's called before the search and
; o  makes the search options uppercase.
; o  removes the M option from the default search options when the current
;    file is not marked.
; o  removes T and B options from deafult search options when a search is
;    executed by the Search dialog. (The Top, Bottom and Cursor radio
;    buttons were not automatically set or unset subject to the T and B
;    default options. Instead of that, they keep their last selection the
;    user has made via the dialog, while after an EPM restart, they open
;    always with the Cursor button activated.)
; o  combines the default search options with the user search options.
; o  removes doubled and exclusive options. (For xcom l, the last option
;    wins.)
; Returns: SearchOptions
defproc ProcessSearchOptions( UserOptions)
   universal default_search_options

   -- Set DefaultOptions to uppercase
   DefaultOptions = upcase( default_search_options)

   -- Switch to All if Mark is default and no text marked, but don't disable
   -- 'M' from UserOptions to make it work as expected.
   if not FileIsMarked() then
      -- Remove 'M' from DefaultOptions
      do forever
         pv = verify( DefaultOptions, 'M', 'M')  -- 2nd arg is charlist to find
         if pv = 0 then
            leave
         endif
         DefaultOptions = delstr( DefaultOptions, pv, 1)
      enddo
   endif

   -- Workaround used when called by the search dialog:
   -- Remove 'T' and 'B' from DefaultOptions, because searchdlg won't
   -- set these checkboxes. Other options are recognized. This is useful
   -- to avoid confusion, because searchdlg will call either defc locate
   -- or defc change. These commands will add default_search_options,
   -- even T or B, while the user would think, he hasn't selected them.
   -- TODO: Better set 'Top' or 'Bottom' checkboxes when opening the
   --       search dialog or when changing default options.
   -- TODO: Automatically set the 'Marked area' checkbox
   --       -  (if mark is visible in the window or)
   --       -  if current file is marked.
   --       Add an option for that as well.
   if pos( 'D', upcase( UserOptions)) then  -- if called from SearchDlg
      do forever
         pv = verify( upcase( DefaultOptions), 'TB', 'M')
         if pv = 0 then
            leave
         endif
         DefaultOptions = delstr( DefaultOptions, pv, 1)
      enddo
   endif

   -- Remove 'T', 'B' and 'U' from DefaultOptions if the new option 'U' is
   -- used. This new option can be added to tell the locate command to remove
   -- 'T' and 'B'. E.g. the 'All' cmd doesn't work with B' or 'T'.
   if pos( 'U', upcase( UserOptions)) then  -- if 'U' is used
      -- Remove 'U' from UserOptions
      do forever
         pv = verify( upcase( UserOptions), 'U', 'M')
         if pv = 0 then
            leave
         endif
         UserOptions = delstr( UserOptions, pv, 1)
      enddo
      -- Remove 'T' and 'B' from DefaultOptions
      do forever
         pv = verify( DefaultOptions, 'TB', 'M')
         if pv = 0 then
            leave
         endif
         DefaultOptions = delstr( DefaultOptions, pv, 1)
      enddo
   endif

   -- Build SearchOptions. The last option wins.
   -- Insert default_search_options just before supplied options (if any)
   -- so the supplied options will take precedence.
   SearchOptions = upcase( DefaultOptions''UserOptions)

   -- Append 'N' to give a message how many changes, if 'Q' not specified
   -- and if all should be changed.
   if pos( '*', SearchOptions) &  -- if e.g. called from SearchDlg
      not pos( 'Q', SearchOptions) then
      SearchOptions = SearchOptions'N'
   endif

   -- Normalize options string and filter out multiple and excluding options
   OptionsList = '+- FR BT AM EC GXW ^ * K N D'    -- for every word in this list: every char excludes each other
   -- Other options: '^ * K N D'
   rest = OptionsList
   SavedSearchOptions = SearchOptions
   SearchOptions = ''
   do w = 1 to words( OptionsList)
      -- Find CurSearchOption
      wrd = word( OptionsList, w)
      restopt = wrd
      CurSearchOption = ''
      CurSearchOptionPos = 0
      do n = 1 to length( restopt)
         parse value restopt with nextopt 2 restopt  -- parse 1 char of restopt
         -- The last option out of wrd in SavedSearchOptions wins
         lp = lastpos( nextopt, SavedSearchOptions)
         if lp > 0 & lp > CurSearchOptionPos then
            CurSearchOption = nextopt
            CurSearchOptionPos = lp
         endif
      enddo
      -- Append to SearchOptions
      SearchOptions = SearchOptions''CurSearchOption
   enddo

   return SearchOptions

; ---------------------------------------------------------------------------
; Preprocessor 2 for Locate and Change. It's called before the search and
; o  determines if a find-next is executed (when search options, search
;    string and file id haven't changed).
; o  removes T and B options when a find-next is executed. (Also the Search
;    dialog doesn't switch automatically from Top or Bottom to Cursor then.)
; o  corrects the cursor position before a search action, when required as a
;    workaround for 'xcom l' bugs to avoid finding false strings.
; Returns: SearchOptions
;          or -1 if a search can't be processed, due to the cursor position
;                and the search direction
defproc ProsessSearchPos( SearchOptions, LastSearchOptions,
                          SearchString, LastSearchString,
                          fid, SearchMode, var fFindNext)
   universal lastsearchpos
   --dprintf( '   ProcessSearchPos: SearchOptions = 'SearchOptions', LastSearchOptions = 'LastSearchOptions)

   parse value lastsearchpos with LastLine LastCol LastFid LastSearchLen LastSearchMode LastSearchRc
   fForeward = lastpos( 'F', SearchOptions) > lastpos( 'R', SearchOptions)
   fDownward = lastpos( '+', SearchOptions) > lastpos( '-', SearchOptions)

   fCursorPosChanged = 1
   do once = 1 to 1
      if LastLine <> .line then
         leave
      elseif LastCol <> .col then
         leave
      endif
      fCursorPosChanged = 0
   enddo

   fOnlyDirectionChanged = 0
   do once = 1 to 1
      -- Remove FR+-D from LastOpts and NextOpts for comparison
      -- with LastSearchOptions and SearchOptions
      LastOpts = LastSearchOptions
      do forever
         pv = verify( LastOpts, 'FR+-D', 'M')
         if pv = 0 then
            leave
         endif
         LastOpts = delstr( LastOpts, pv, 1)
      enddo
      NextOpts = SearchOptions
      do forever
         pv = verify( NextOpts, 'FR+-D', 'M')
         if pv = 0 then
            leave
         endif
         NextOpts = delstr( NextOpts, pv, 1)
      enddo

      if fid <> LastFid then
         leave
      elseif LastSearchString <> SearchString then
         leave
      elseif LastSearchOptions <> SearchOptions then
         if LastOpts = NextOpts then
            -- This will set fFindNext = 1 and then fMoveCursor = 1
            fOnlyDirectionChanged = 1
         else
            leave
         endif
      endif
   enddo

   OptsTBRemoved = SearchOptions
   do forever
      pv = verify( OptsTBRemoved, 'TB', 'M')
      if pv = 0 then
         leave
      endif
      OptsTBRemoved = delstr( OptsTBRemoved, pv, 1)
   enddo
   LastOptsTBRemoved = LastSearchOptions
   do forever
      pv = verify( LastOptsTBRemoved, 'TB', 'M')
      if pv = 0 then
         leave
      endif
      LastOptsTBRemoved = delstr( LastOptsTBRemoved, pv, 1)
   enddo

   fFindNext = 0
   do once = 1 to 1
      if fid <> LastFid then
         leave
      elseif LastSearchRc <> 0 then
         leave
      elseif fCursorPosChanged then
         leave
      elseif LastSearchString <> SearchString then
         leave
      elseif OptsTBRemoved <> LastOptsTBRemoved & not fOnlyDirectionChanged then
         leave
      else
         fFindNext = 1
      endif
   enddo

   do once = 1 to 1
      -- Disable further search if the previous search
      -- with the same fid, SearchString, SearchOptions and SearchPos wasn't
      -- successfull. This is required for T and B to prevent looping after a
      -- single search stop at the end.
      if fid <> LastFid then
         leave
      elseif LastSearchString <> SearchString then
         leave
      elseif fCursorPosChanged then
         leave
      elseif LastSearchRc = 0 then
         leave
      elseif fOnlyDirectionChanged then
         leave
      --elseif OptsTBRemoved <> LastOptsTBRemoved then
      elseif SearchOptions <> LastSearchOptions then
         leave
      else
         return -1  -- -1 means: don't search
      endif
   enddo

   -- FindNext: Remove 'T' and 'B' for the search execution (not from LastSearchArgs)
   FilteredSearchOptions = SearchOptions
   if fFindNext then
      --dprintf( '   ProcessSearchPos: Removing T and B')
      FilteredSearchOptions = OptsTBRemoved
   endif

   -- FindNext or backward: Move cursor first to not find the just found string again
   fMoveCursor = 0
   do once = 1 to 1

      -- Change should process the string at cursor. The cursor pos may result
      -- from a previous locate.
      -- If there is no string to change at the cursor, it processes the next
      -- string. For that no cursor move is required.
      if LastSearchMode = 'l' & SearchMode = 'c' then
         --dprintf( '   ProcessSearchPos: Search mode: l -> c ==> leave')
         leave

      -- If the cursor was moved manually after the last search, don't move it
      -- before the next search.
      elseif fCursorPosChanged then
         leave

      -- FindNext requires to move the cursor first, before the next search.
      -- Reverse search requires to move the cursor first, before the next search.
      -- Bug in xcom l: reverse search finds also string right of cursor
      -- Move cursor to not find the just found string again
      elseif not fFindNext & fForeward then
         leave
      endif

      fMoveCursor = 1
   enddo
   --dprintf( '   ProcessSearchPos: fCursorPosChanged = 'fCursorPosChanged', fOnlyDirectionChanged = 'fOnlyDirectionChanged', fMoveCursor = 'fMoveCursor', fFindNext = 'fFindNext)

   fSearch = 1
   -- Move cursor to not find the string at cursor (again)
   if fMoveCursor then
      if LastSearchLen = '' then
         LastSearchLen = 0
      endif
      if fForeward then  -- must be a FindNext
         -- Foreward
         if fOnlyDirectionChanged & fFindNext & LastSearchRc <> 0 then
            next = .col
         else
            -- Foreward: move amount of LastSearchLen right
            next = .col + LastSearchLen
         endif
         if next >= length( textline( .line)) + 1 then
            if fDownward then
               if .line < .last then
                  down
                  .col = 1
               else
                  fSearch = 0  -- can't move down at the bottom
               endif
            else
               if .line > 1 then
                  up
                  .col = 1
               else
                  fSearch = 0  -- can't move up at the top
               endif
            endif
         else
            .col = next
         endif
      else
         -- Backward: move 1 char left
         next = .col - 1
         if next < 1 then
            if fDownward then
               if .line < .last then
                  down
                  end_line
               else
                  fSearch = 0  -- can't move down at the bottom
               endif
            else
               if .line > 1 then
                  up
                  end_line
               else
                  fSearch = 0  -- can't move up at the top
               endif
            endif
         else
            .col = next
         endif
      endif
   endif

   -- Check if cursor is after or before the mark
   getmark markfirstline, marklastline, markfirstcol, marklastcol, markfid
   if pos( 'M', upcase( SearchOptions)) then
      if fDownward then
         if not pos( 'T', upcase( SearchOptions)) then
            if .line > marklastline then
               fSearch = 0
            endif
         endif
      else
         if not pos( 'B', upcase( SearchOptions)) then
            if .line < markfirstline then
               fSearch = 0
            endif
         endif
      endif
      if fForeward then
         if not pos( 'T', upcase( SearchOptions)) then
            if .line = marklastline then
               if .col > marklastcol then
                  fSearch = 0
               endif
            endif
         endif
      else
         if not pos( 'B', upcase( SearchOptions)) then
            if .line = markfirstline then
               if .col < markfirstcol then
                  fSearch = 0
               endif
            endif
         endif
      endif
   endif

   if not fSearch then
      return  -1
   else
      return FilteredSearchOptions
   endif

; ---------------------------------------------------------------------------
; Preprocessor 3 for Locate and Change. It's called after a successfull
; search within a mark and
; o  recognizes if a search within a mark is restarted (and should rather be
;    stopped)
; o  checks if found string is completely within the mark (required for
;    char mark only)
; Returns: rc = -273 if search has restarted
;               -280 if found string is not completely within the mark
;                  0 otherwise
defproc ProsessSearchMark( SearchOptions, LastSearchOptions,
                           SearchString, LastSearchString,
                           fFindNext)
   universal lastsearchpos

   -- Search in mark: disable search restart when at end
   -- Bug in any mark search: At the end of the mark the search loops to
   -- the begin of the mark. Mark search won't stop at the mark end.
   parse value lastsearchpos with LastLine LastCol LastFid LastSearchLen LastSearchMode LastSearchRc
   fForeward = lastpos( 'F', SearchOptions) > lastpos( 'R', SearchOptions)
   fDownward = lastpos( '+', SearchOptions) > lastpos( '-', SearchOptions)

   fRestarted = 0
   if fFindNext then
      if fDownward then
         if .line < LastLine then
            fRestarted = 1
         endif
      else
         if .line > LastLine then
            fRestarted = 1
         endif
      endif
      if .line = LastLine then
         if fForeward then
            if .col < LastCol then
               fRestarted = 1
            endif
         else
            if .col > LastCol then
               fRestarted = 1
            endif
         endif
      endif
   endif

   -- Search in char mark: repeat search if found string is not full in mark
   -- Bug for char mark search: The first and last marked lines were completely
   -- searched, like for a line mark.
   fRepeat = 0
   markt = marktype()
   if markt = 'CHAR' then
      SearchStringLen = GetPmInfo( EPMINFO_LSLENGTH)
      if not InMark( .line, .col, SearchStringLen) then
         fRepeat = 1
      endif
   endif

   if fRestarted then
      return -273  -- String not found
   elseif fRepeat then
      return -280  -- Text not marked
   else
      return 0
   endif

; ---------------------------------------------------------------------------
; LastSearchArgs and LastChangeArgs are no longer universals. To make them
; global across all EPM windows, they are saved in NEPMD.INI only.
; ---------------------------------------------------------------------------
defproc GetLastSearchArgs
   KeyPath = '\NEPMD\Var\Search\LastSearchArgs'
   LastSearchArgs = QueryConfigKey( KeyPath)
   --dprintf( 'GetLastSearchArgs: LastSearchArgs = 'LastSearchArgs)
   return LastSearchArgs

; ---------------------------------------------------------------------------
defproc GetLastChangeArgs
   KeyPath = '\NEPMD\Var\Search\LastChangeArgs'
   LastChangeArgs = QueryConfigKey( KeyPath)
   return LastChangeArgs

; ---------------------------------------------------------------------------
defproc GetLastSearchDialogOptions
   KeyPath = '\NEPMD\Var\Search\LastSearchDialogOptions'
   LastSearchDialogOptions = QueryConfigKey( KeyPath)
   --dprintf( 'GetLastSearchDialogOptions: LastSearchDialogOptions = 'LastSearchDialogOptions)
   return LastSearchDialogOptions

; ---------------------------------------------------------------------------
defproc SetLastSearchArgs
   KeyPath = '\NEPMD\Var\Search\LastSearchArgs'
   LastSearchArgs = arg( 1)
   call WriteConfigKey( KeyPath, LastSearchArgs)
   --dprintf( 'SetLastSearchArgs: LastSearchArgs = 'LastSearchArgs)
   return

; ---------------------------------------------------------------------------
defproc SetLastChangeArgs
   KeyPath = '\NEPMD\Var\Search\LastChangeArgs'
   LastChangeArgs = arg( 1)
   call WriteConfigKey( KeyPath, LastChangeArgs)
   return

; ---------------------------------------------------------------------------
defproc SetLastSearchDialogOptions
   KeyPath = '\NEPMD\Var\Search\LastSearchDialogOptions'
   LastSearchDialogOptions = arg( 1)
   call WriteConfigKey( KeyPath, LastSearchDialogOptions)
   --dprintf( 'SetLastSearchDialogOptions: LastSearchDialogOptions = 'LastSearchDialogOptions)
   return

; ---------------------------------------------------------------------------
defproc GetAddSearchChangeOptions
   KeyPath = '\NEPMD\Var\Search\AddSearchChangeOptions'
   AddSearchChangeOptions = QueryConfigKey( KeyPath)
   return AddSearchChangeOptions

; ---------------------------------------------------------------------------
defproc SetAddSearchChangeOptions
   KeyPath = '\NEPMD\Var\Search\AddSearchChangeOptions'
   AddSearchChangeOptions = arg( 1)
   call WriteConfigKey( KeyPath, AddSearchChangeOptions)
   return

; ---------------------------------------------------------------------------
; Syntax: locate !<SearchString>[!<UserOptions>]
;         The first char will be taken as delimitter, in this case '!'.
;
; New: Without args, a FindNext will be executed.
;
; Added a new search option: 'U'. This will replace any occurance of
; 'T' and 'B' in default_search_options for this locate command.
;
; Note:  this DEFC also gets executed by the slash ('/') command and by the
; search dialog. The search dialog adds option 'D'.
defc RepeatFind, L, Locate
   universal default_search_options
   universal lastsearchpos

   sayerror 0  -- delete previous message from messageline
   LastSearchArgs = GetLastSearchArgs()
   LastChangeArgs = GetLastChangeArgs()
   LastSearchDialogOptions = GetLastSearchDialogOptions()

   args = strip( arg( 1), 'L')

   -- Filter out 'D' arg from search dialog for RepeatFind
   fDialogRepeat = 0
   if args = \1\1\1\1'D' then
      fDialogRepeat = 1
      args = ''
   endif

   -- Handle RepeatFind
   if args = '' then  -- If no args, query args
      SearchArgs = LastSearchArgs

      tmpdelim = substr( SearchArgs, 1, 1)  -- get 1st delimiter
      parse value SearchArgs with (tmpdelim)SearchString(tmpdelim)SearchOptions
      next = SearchOptions

      -- Ensure that option D is removed if not coming from the search dialog
      if not fDialogRepeat then
         do forever
            Dpos = pos( 'D', next)
            if Dpos = 0 then
               leave
            endif
            next = delstr( next, Dpos, 1)
         enddo
      endif
      -- Ensure that last options from the dialog are used, not from elsewhere
      if fDialogRepeat then
         next = LastSearchDialogOptions
      endif

      -- Rebuild SearchArgs
      SearchOptions = next
      SearchArgs = tmpdelim''SearchString''tmpdelim''SearchOptions

      args = SearchArgs
      -- Process the parsing of args again to recognize a possible change of
      -- default_search_options in the meantime.
   endif

   delim = substr( args, 1, 1)  -- get 1st delimiter
   parse value args with (delim)SearchString(delim)UserOptions
   UserOptions = strip( UserOptions, 'T', delim)

   AddSearchChangeOptions = GetAddSearchChangeOptions()
   UserOptions = UserOptions''AddSearchChangeOptions

   tmpdelim = substr( LastSearchArgs, 1, 1)  -- get 1st delimiter
   parse value LastSearchArgs with (tmpdelim)LastSearchString(tmpdelim)LastSearchOptions

   -- The current SearchString is compared with that of LastChangeArgs.
   -- If another string was specified, then LastChangeArgs is reset to ''.
   tmpdelim = substr( LastChangeArgs, 1, 1)  -- get 1st delimiter
   parse value LastChangeArgs with (tmpdelim)cSearchString(tmpdelim)cReplaceString(tmpdelim)cSearchOptions

   getfileid fid

   -- Prepend default options and normalize search options
   SearchOptions = ProcessSearchOptions( UserOptions)

   -- Build list of search args with options, last option wins.
   -- Save the universal var here. Later ProsessSearchPos changes
   -- SearchOptions if required.
   SearchArgs = delim''SearchString''delim''SearchOptions
   --dprintf( overlay( ' Locate ', copies( '-', 60), 4))
   --dprintf( '--- SearchArgs = 'SearchArgs', arg( 1) = 'arg( 1)', fDialogRepeat = 'fDialogRepeat)
   --dprintf( 'LastSearchArgs = 'LastSearchArgs', LastSearchDialogOptions = 'LastSearchDialogOptions)

   parse value lastsearchpos with LastLine LastCol LastFid LastSearchLen LastSearchMode LastSearchRc

   SearchMode = 'l'

   --dprintf( 'Before ProsessSearchPos: SearchOptions = 'SearchOptions', LastSearchOptions = 'LastSearchOptions)
   -- Maybe move cursor and remove T and B search options for a FindNext
   fFindNext = 0
   FilteredSearchOptions = ProsessSearchPos( SearchOptions, LastSearchOptions,
                                             SearchString, LastSearchString,
                                             fid, SearchMode,
                                             fFindNext)  -- fFindNext is set
   --dprintf( 'After ProsessSearchPos:  FilteredSearchOptions = 'FilteredSearchOptions', LastSearchOptions = 'LastSearchOptions)
   fSearch = 1
   if FilteredSearchOptions = -1 then
      -- Omit search at the top or at the bottom
      fSearch = 0
   else
      -- SearchOptions may be changed by ProcessSearchPos
      SearchArgs         = delim''SearchString''delim''SearchOptions
      FilteredSearchArgs = delim''SearchString''delim''FilteredSearchOptions
   endif

   -- Save last and cur args
   call SetLastSearchArgs( SearchArgs)
   if pos( 'D', SearchOptions) then
      call SetLastSearchDialogOptions( SearchOptions)
   endif

   if SearchString = cSearchString then
      -- Reset the SearchOptions in LastChangeArgs for the next Change while
      -- keeping the previous ReplaceString.
      --dprintf( 'Change LastChangeArgs to use the options from Locate')
      ChangeArgs = delim''cSearchString''delim''cReplaceString''delim''SearchOptions
      call SetLastChangeArgs( ChangeArgs)
   else
      -- Reset LastChangeArgs when a new search string was specified.
      call SetLastChangeArgs( '')
   endif

   -- Pos from before the search
   call pSave_Pos( savedpos)
   fRestorePos = 1

   lrc = -273  -- String not found
   if fSearch then
      display -8  -- suppress writing to MsgBox
      'xcom l 'FilteredSearchArgs
      lrc = rc
      call pSave_Pos( lpos)
      display 8
   endif
   SetAddSearchChangeOptions( '')

   fRepeat = 0
   if fSearch & lrc = 0 & pos( 'M', SearchOptions) > 0 then
      -- Check if found string is fully marked (if not: rc = -280) and
      -- if Search is not looping within the mark (if: rc = -273)
      SearchMarkRc = ProsessSearchMark( FilteredSearchOptions, LastSearchOptions,
                                        SearchString, LastSearchString,
                                        fFindNext)
      if SearchMarkRc = -273 then  -- String not found
         lrc = SearchMarkRc
         fRepeat = 0
         fRestorePos = 1
         fSearch = 0
      elseif SearchMarkRc = -280 then  -- Text not marked
         lrc = SearchMarkRc
         fRepeat = 1
      endif
   endif

   if not fRepeat then
      -- Give error message if search was omitted
      if not fSearch then
         display -8  -- suppress writing to MsgBox
         sayerror sayerrortext( lrc)  -- The same as: 'SayError -273'
         display 8
      endif

      -- Highlight it and maybe scroll to cursor pos
      if lrc = 0 then
         fRestorePos = 0
         -- Cursor position here must be that from before the search (savedpos)
         call ScrollAfterLocate( savedpos, lpos)
         -- SearchLen will be queried by getpminfo( EPMINFO_LSLENGTH)
         call Highlight_Match()
      endif

      if fRestorePos then
         call pRestore_Pos( savedpos)
         .line = .line           -- maybe scroll to ensure that cursor is visible
         'HighlightCursor'
      endif
   endif

   -- Save last searched pos, file, len, search mode and rc
   -- For a repeated search, the start pos. will be moved by len to avoid
   -- finding the same string multiple times.
   -- GetPmInfo queries the length right, also for the grep or egrep search.
   len  = GetPmInfo( EPMINFO_LSLENGTH)
   --dprintf( 'len = 'len', linelen = 'length( textline( .line))', xcom l 'SearchArgs)
   if (len = 0 | len = '') & (lrc = 0 | fRepeat) then
      len = length( SearchString)
   endif
   thissearchpos = .line .col fid len SearchMode lrc
   --dprintf( 'thissearchpos = 'thissearchpos', lastsearchpos = 'lastsearchpos)
   lastsearchpos = thissearchpos

   rc = lrc  -- does hightlight_match change rc?

   if fRepeat then
      -- Replace the centered scroll pos of 'xcom l'.
      -- Cursor position here must be that from before the search (savedpos)
      call ScrollAfterLocate( savedpos, lpos)

      -- The position must be advanced before the next search.
      -- RepeatFind does that by checking lastsearchpos in ProsessSearchPos.
      'RepeatFind'
   endif

; ---------------------------------------------------------------------------
defc RepeatChange, C, Change
   universal default_search_options
   universal lastsearchpos
   universal stay  -- if 1, then restore pos even after a successful change

   sayerror 0  -- delete previous message from messageline
   LastSearchArgs = GetLastSearchArgs()
   LastChangeArgs = GetLastChangeArgs()
   LastSearchDialogOptions = GetLastSearchDialogOptions()

   args = strip( arg( 1), 'L')

   -- Filter out 'D' arg from search dialog for RepeatChange
   fDialogRepeat = 0
   if args = \1\1\1\1'D' then
      fDialogRepeat = 1
      args = ''
   endif

   -- Handle RepeatChange
   if args = '' then   -- If no args, query lastchangeargs
      ChangeArgs = LastChangeArgs

      tmpdelim = substr( ChangeArgs, 1, 1)  -- get 1st delimiter
      parse value ChangeArgs with (tmpdelim)SearchString(tmpdelim)ReplaceString(tmpdelim)SearchOptions
      next = SearchOptions

      -- Ensure that option D is removed if not coming from the search dialog
      if not fDialogRepeat then
         do forever
            Dpos = pos( 'D', next)
            if Dpos = 0 then
               leave
            endif
            next = delstr( next, Dpos, 1)
         enddo
      endif
      -- Ensure that last options from the dialog are used, not from elsewhere
      if fDialogRepeat then
         next = LastSearchDialogOptions
      endif

      -- Rebuild LastChangeArgs
      SearchOptions = next
      ChangeArgs = tmpdelim''SearchString''tmpdelim''ReplaceString''tmpdelim''SearchOptions

      args = ChangeArgs
      -- Process the parsing of args again to recognize a possible change of
      -- default_search_options in the meantime.
   endif

   delim = substr( args, 1, 1)  -- get 1st delimiter
   p2 = pos( delim, args, 2)    -- check 2nd delimiter of 2 or 3
   if not p2 then
      sayerror NO_REP__MSG  -- 'No replacement string specified' (Rather the missing delimiter is checked.)
      -- A defc must use "return myrc" or use "rc = myrc" to set the
      -- global var rc. When "return" is used only, rc would be set to empty.
      -- Another option is to omit return for a defc. Then rc will be kept as well.
      -- A defproc may return its own rcx, without overriding the global rc.
      rc = -257  -- Invalid number of parameters
      return rc
   endif
   parse value args with (delim)SearchString(delim)ReplaceString(delim)UserOptions
   UserOptions = strip( UserOptions, 'T', delim)

   AddSearchChangeOptions = GetAddSearchChangeOptions()
   UserOptions = UserOptions''AddSearchChangeOptions

   tmpdelim = substr( LastSearchArgs, 1, 1)  -- get 1st delimiter
   parse value LastSearchArgs with (tmpdelim)LastSearchString(tmpdelim)LastSearchOptions

   -- If LastChangeArgs was reset by the last search ...
   if LastChangeArgs = '' then
      -- ... get args from last search, not from last change
      LastChangeArgs = delim''LastSearchString''delim''ReplaceString''delim''LastSearchOptions
   endif
   tmpdelim = substr( LastChangeArgs, 1, 1)  -- get 1st delimiter
   parse value LastChangeArgs with (tmpdelim)LastSearchString(tmpdelim)LastReplaceString(tmpdelim)LastSearchOptions

   getfileid fid

   -- Prepend default options and normalize search options
   SearchOptions = ProcessSearchOptions( UserOptions)

   -- Build list of change args with options, last option wins.
   -- Save the universal var here. Later ProsessSearchPos changes
   -- SearchOptions if required.
   ChangeArgs = delim''SearchString''delim''ReplaceString''delim''SearchOptions
   -- Set LastSearchArgs as well, to use first Ctrl+F and then Ctrl+C for to
   -- operate on the same SearchString. Even a ChangeNext should synchronize it.
   SearchArgs = delim''SearchString''delim''SearchOptions
   --dprintf( overlay( ' Change ', copies( '-', 60), 4))
   --dprintf( '--- ChangeArgs = 'ChangeArgs', arg( 1) = 'arg( 1)', fDialogRepeat = 'fDialogRepeat)
   --dprintf( '--- SearchArgs = 'SearchArgs', arg( 1) = 'arg( 1))
   --dprintf( 'LastChangeArgs = 'LastChangeArgs', LastSearchArgs = 'LastSearchArgs', LastSearchDialogOptions = 'LastSearchDialogOptions)

   -- Save last args
   call SetLastChangeArgs( ChangeArgs)
   call SetLastSearchArgs( SearchArgs)
   if pos( 'D', SearchOptions) then
      call SetLastSearchDialogOptions( SearchOptions)
   endif

   SearchMode = 'c'

   --dprintf( 'Before ProsessSearchPos: SearchOptions = 'SearchOptions', LastSearchOptions = 'LastSearchOptions)
   -- Maybe move cursor and remove T and B search options for a FindNext
   fFindNext = 0
   FilteredSearchOptions = ProsessSearchPos( SearchOptions, LastSearchOptions,
                                             SearchString, LastSearchString,
                                             fid, SearchMode,
                                             fFindNext)  -- fFindNext is set
   --dprintf( 'After ProsessSearchPos:  FilteredSearchOptions = 'FilteredSearchOptions', LastSearchOptions = 'LastSearchOptions)

   fSearch = 1
   if FilteredSearchOptions = -1 then  -- never true for 'change'
      -- Omit search at the top or at the bottom
      fSearch = 0
   else
      -- SearchOptions may be changed by ProsessSearchPos
      FilteredChangeArgs = delim''SearchString''delim''ReplaceString''delim''FilteredSearchOptions
   endif

   -- Pos from before the search
   call pSave_Pos( savedpos)
   fRestorePos = 1

   OldLineLen = 0
   NewLineLen = 0
   if fSearch then
      if pos( 'G', SearchOptions) > 0 | pos( 'X', SearchOptions) > 0 then
         -- For grep and egrep searches only:
         -- Also build SearchArgs for finding the same line where Change will process
         -- next. This allows for querying the line length before the change and
         -- for calculating the change length.
         TmpSearchArgs = delim''SearchString''delim''FilteredSearchOptions
         display -8
         'xcom l 'TmpSearchArgs
         if rc = 0 then
            OldLineLen = length( textline( .line))
            --dprintf( 'OldLineLen = 'OldLineLen', .line = '.line)
         endif
         call pRestore_Pos( savedpos)
         display 8
      endif
   endif

   lrc = -273  -- String not found
   if fSearch then
      call NextCmdAltersText()
      display -8
      'xcom c 'FilteredChangeArgs
      lrc = rc
      NewLineLen = length( textline( .line))
      call pSave_Pos( lpos)
      display 8
   endif
   SetAddSearchChangeOptions( '')

   fRepeat = 0
   if fSearch & lrc = 0 & pos( 'M', SearchOptions) > 0 then
      -- Check if found string is fully marked (if not: rc = -280) and
      -- if Search is not looping within the mark (if: rc = -273)
      SearchMarkRc = ProsessSearchMark( SearchOptions, LastSearchOptions,
                                        SearchString, LastSearchString,
                                        fFindNext)
      if SearchMarkRc = -273 then  -- String not found
         lrc = SearchMarkRc
         fRepeat = 0
         fRestorePos = 1
         fSearch = 0
      elseif SearchMarkRc = -280 then  -- Text not marked
         lrc = SearchMarkRc
         fRepeat = 1
      endif
   endif

   -- Save last searched pos, file, len, search mode and rc
   -- For a repeated search, the start pos. will be moved by len to avoid
   -- finding the same string multiple times.
   -- GetPmInfo queries the length right, also for the grep or egrep search.
   -- SearchLen can be queried for a Search action by getpminfo( EPMINFO_LSLENGTH).
   -- But getpminfo( EPMINFO_LSLENGTH) gives the value for the search string,
   -- not for the change string.
   -- While for a non-grep search the length of the replacement string
   -- is known, it has to be calculated with the line length from before
   -- the change for a grep search.
   if pos( 'G', SearchOptions) > 0 | pos( 'X', SearchOptions) > 0 then
      OldLen = getpminfo( EPMINFO_LSLENGTH)
      if OldLen = '' then
         OldLen = 0
      endif
      NewLen = OldLen - OldLineLen + NewLineLen
      --dprintf( 'NewLen = 'NewLen)
      ReplaceLen = NewLen
   else
      ReplaceLen = length( ReplaceString)
   endif
   thissearchpos = .line .col fid ReplaceLen SearchMode lrc
   --dprintf( 'thissearchpos = 'thissearchpos', lastsearchpos = 'lastsearchpos)
   lastsearchpos = thissearchpos

   if not fRepeat then
      -- Give error message if search was omitted
      if not fSearch then
         display -8  -- suppress writing to MsgBox
         sayerror sayerrortext( lrc)  -- The same as: 'SayError -273'
         display 8
      endif

      -- Highlight it and maybe scroll to cursor pos
      if lrc = 0 then
         fRestorePos = 0
         -- Cursor position here must be that from before the search (savedpos)
         call ScrollAfterLocate( savedpos, lpos)

         --call highlight_match()  -- gives wrong value
         call Highlight_Match( .line .col ReplaceLen)

         -- Restore pos after change command if stay = 1
         if stay then
            call pRestore_Pos( savedpos)
         endif
      endif

      if fRestorePos then
         call pRestore_Pos( savedpos)
         .line = .line           -- maybe scroll to ensure that cursor is visible
         'HighlightCursor'
      endif
   endif

   rc = lrc  -- does hightlight_match change rc?

   if fRepeat then
      -- Replace the centered scroll pos of 'xcom c'.
      -- Cursor position here must be that from before the search (savedpos)
      call ScrollAfterLocate( savedpos, lpos)

      -- The position must be advanced before the next search.
      -- RepeatFind does that by checking lastsearchpos in ProsessSearchPos.
      'RepeatFind'
   endif

; ---------------------------------------------------------------------------
; Highlight a "hit" after a Locate command or Repeat_find operation.
; Never used its previous arg( 1) = search_len in 6.03b.
; New: optional arg( 1) = <line> <col> <len>
defproc Highlight_Match

   if rc then  -- if not found; rc was set from last 'c'|'l'|repeat_find
      return
   endif
   savedrc = rc

   parse arg args
   if args <> '' then
      parse arg line col len
      if col = '' then
         -- This must be the previously used syntax: arg( 1) = search_len
         args = ''
      endif
   endif
   if args = '' then
      line = .line
      col  = getpminfo( EPMINFO_SEARCHPOS)
      len  = getpminfo( EPMINFO_LSLENGTH)
   endif
   if col = 0 then
      col = .col
   endif

   -- Draw a circle around the found string
   circleit LOCATE_CIRCLE_STYLE,
      line,
      col,
      col + len - 1,
      LOCATE_CIRCLE_COLOR1,
      LOCATE_CIRCLE_COLOR2

   rc = savedrc
   return

; ---------------------------------------------------------------------------
; Callable with 'postme'. Required for GlobalFind.
defc HighlightMatch
   call Highlight_Match( arg( 1))

; ---------------------------------------------------------------------------
defc CircleIt
   parse arg line startcol endcol
   circleit LOCATE_CIRCLE_STYLE, line, startcol, endcol,
            LOCATE_CIRCLE_COLOR1, LOCATE_CIRCLE_COLOR2

; ---------------------------------------------------------------------------
; Syntax: ScrollAfterLocate( savedpos [, lpos] [, UpDown] [, ScrollSpace])
;   savedpos       = pos before the search
;   lpos           = pos after the search, at the found string (optional)
;   ScrollToTopBot = 'T' or 'B' (optional)
;   ScrollSpace    = space or window line values as 'top bot' (optional)
;                    Negative values are handled as vertical space,
;                    positive values are handled as window height line number,
;                    where the topmost visible line is 1 and the bottommost
;                    visible line is .windowheight.
;                    Example: '26 -4' means:
;                       topmost window line to scroll to = 26
;                          (the vertical space between 1 and 25 can be used for the search dialog)
;                       bottom minimum space = 4 lines
;                    If possible, vertical window scrolling is avoided. Then the loacate
;                    pos must be within the specified scroll area.
defproc ScrollAfterLocate
   savedpos = arg( 1)
   if savedpos = '' then
      sayerror 'Error in ScrollAfterLoacate call: 'sayerrortext( -268)'. savedpos is missing.'
      return -268
   else
      parse value savedpos with svline svcol svcx svcy .
   endif

   lpos = arg( 2)
   if lpos = '' then
      lline = .line
      lcol = .col
      lcx = .cursorx
      lcy = .cursory
   else
      parse value lpos with lline lcol lcx lcy .
   endif

   ScrollToTopBot = upcase( arg( 3))
   if ScrollToTopBot = 'T' then
      -- Assist: inverted arg to sroll to the bottom after a downward search
      fDownward = 1
   elseif ScrollToTopBot = 'B' then
      -- Assist: inverted arg to sroll to the top after an upward search
      fDownward = 0
   else
      -- Get search direction from previous search options
      getsearch CurSearch
      parse value CurSearch with 'xcom' SearchMode rest
      rest = strip( rest)
      delim = leftstr( rest, 1)
      if SearchMode = 'c' then
         parse value rest with (delim)SeachString(delim)ReplaceString(delim)SearchOptions
      else
         parse value rest with (delim)SeachString(delim)SearchOptions
      endif
      fDownward = lastpos( '+', SearchOptions) >= lastpos( '-', SearchOptions)
   endif

   ScrollSpace = arg( 4)
   if ScrollSpace = '' then
      ScrollSpace = QueryConfigKey( '\NEPMD\User\CursorPos\ScrollAfterLocate')
   endif

   topline  = svline - svcy + 1
   botline  = svline - svcy + .windowheight
   leftcol  = svcol  - svcx + 1
   rightcol = svcol  - svcx + .windowwidth

   topheight = word( ScrollSpace, 1)
   botheight = word( ScrollSpace, 2)
   if not IsNum( topheight) then
      topheight = 4
   elseif topheight < 0 then
      -- Minus: topheight is vspace in lines from top, set absolute value + 1
      topheight = -1 * topheight + 1
   endif
   if not IsNum( botheight) then
      botheight = .windowheight - 4
   elseif botheight < 0 then
      -- Minus: botheight is vspace in lines from bottom, calc with absolute value
      botheight = .windowheight + botheight
   endif
   -- Ensure that area is at least 1 line height
   if botheight - topheight < 0 then
      topheight = .windowheight % 2
      botheight = topheight
   endif

   fInScrollArea = 0
   x = lcol
   y = lline
   if     y < topline + topheight then
   elseif y > botline - .windowheight + botheight then
   elseif x < leftcol then
   elseif x > rightcol then
   else
      fInScrollArea = 1
   endif

   -- Minimize scrolling for ScrollSpace = '0 0'
   if topheight = 0 & botheight = 0 then
      fInArea = OnScreen()
   -- Minimize scrolling if line number hasn't changed
   elseif lline = svline then
      fInArea = OnScreen()
   else
      fInArea = fInScrollArea
   endif

   if fInArea then
      call pRestore_Pos( savedpos)
      -- Set line to lline without scrolling
      .cursory = svcy + lline - svline
      .col  = lcol
   else
      -- Reposition cursor
      call pRestore_Pos( lpos)
      oldline = .line

      if fDownward then
         .cursory = topheight
      else
         .cursory = botheight
      endif

      .line = oldline
   endif

; ---------------------------------------------------------------------------
; Can also be called with C or F as arg to repeat last change or find.
/*
Ŀ
 what's it called: searchdlg       syntax:   searchdlg [next]               
                                                                            
 what does it do : ask EPM.EXE to pop up its internal search & replace dlg. 
                   This is done by posting a EPM_POPCHANGEDLG message to the
                   EPM Book window.                                         
                   if the [next] param = 'F'  a find next will take place   
                   if the [next] param = 'C'  a change next will take place 
                                                                            
                   (All EPM_EDIT_xxx messages are defined in the ETOOLKT    
                    PACKAGE available on PCTOOLS.)                          
                                                                            
 who and when    : Jerry C.   2/27/89                                       

*/
defc SearchDlg
   universal default_search_options

   parse value upcase( arg( 1)) with uparg .

   if uparg = 'C' then
      'RepeatChange '\1\1\1\1'D'
   elseif uparg = 'F' then
      'RepeatFind '\1\1\1\1'D'

   else
      -- The application will free the buffer allocated by this macro
      call windowmessage( 0,  getpminfo( APP_HANDLE),
                          5128,               -- EPM_POPCHANGEDLG
                          0,
                          Put_In_Buffer( default_search_options))
   endif

; ---------------------------------------------------------------------------
; Returns '+' or '-'.
defproc GetSearchDirection
   universal default_search_options
   ret = '+'

   args = GetLastSearchArgs()
   delim = substr( args, 1, 1)  -- get 1st delimiter
   parse value args with (delim)SearchString(delim)UserOptions
   UserOptions = strip( UserOptions, 'T', delim)

   -- Analyze only last search options, not last change options
   Minuspos = lastpos( '-', default_search_options''UserOptions)
   Pluspos  = lastpos( '+', default_search_options''UserOptions)

   if MinusPos > PlusPos then
      ret = '-'
   endif
   return ret

; ---------------------------------------------------------------------------
; From EPMSMP\REVERSE.E
; Search options for specifying the direction:
;    F  start searching from the start of line
;    R  start searching from the end of line
;    +  start searching from the start of text
;    -  start searching from the end of text
; The relevant options are + and -, F or R is set automatically.
;
; Original by Larry Margolis:
;    Repeat the previous Locate, but in the reverse direction.
;    E.g., if you search for a string that you know exists, but
;    it's not found before the end of the file, press Ctrl+minus
;    to repeat the search for the same string looking from the
;    cursor position to the beginning of the file.
; Changed:
;    Toggles the search direction (options +F or -R) without any
;    following locate action.
;    Doesn't produce an error msg anymore if oldsearch = empty.
; This has no effect on the settings of the Search dialog.
defc ToggleSearchDirection
   universal default_search_options

   args = GetLastSearchArgs()
   delim = substr( args, 1, 1)  -- get 1st delimiter
   parse value args with (delim)SearchString(delim)UserOptions
   UserOptions = strip( UserOptions, 'T', delim)

   -- Analyze only last search options, not last change options
   Minuspos = lastpos( '-', default_search_options''UserOptions)
   Pluspos  = lastpos( '+', default_search_options''UserOptions)

   -- Append +F or -R
   if Minuspos > Pluspos then  -- in searchoptions: the last option wins
      'SearchDirection +'
      'SayHint Changed search direction to: forward.'
   else
      'SearchDirection -'
      'SayHint Changed search direction to: backward.'
   endif

; ---------------------------------------------------------------------------
; Set SearchDirection to foreward (arg = 'F' or '+') or backward (arg = 'R'
; or '-').
defc SearchDirection
   -- Options will be normalized by ProcessSearchOptions. Therefore it's
   -- possible to simply append options. (The last option wins.)
   AddSearchChangeOptions = ''
   Direction = upcase( arg( 1))
   if Direction = '' then
      -- nop
   elseif Direction = '+' | Direction = 'F' then
      AddSearchChangeOptions = '+F'
   elseif Direction = '-' | Direction = 'R' then
      AddSearchChangeOptions = '-R'
   endif
   if AddSearchChangeOptions <> '' then
      SetAddSearchChangeOptions( AddSearchChangeOptions)
   endif
   return

; ---------------------------------------------------------------------------
defc FindNext
   'SearchDirection +'
   'RepeatFind'

; ---------------------------------------------------------------------------
defc FindPrev
   'SearchDirection -'
   'RepeatFind'

; ---------------------------------------------------------------------------
defc ChangeFindNext
   'SearchDirection +'
   'RepeatChange'
   if not rc then
      'RepeatFind'
   endif

; ---------------------------------------------------------------------------
defc ChangeFindPrev
   'SearchDirection -'
   'RepeatChange'
   if not rc then
      'RepeatFind'
   endif

; ---------------------------------------------------------------------------
defc RepeatFindAllFiles, RingFind

   LastSearchArgs = GetLastSearchArgs()
   delim = substr( LastSearchArgs, 1, 1)  -- get 1st delimiter
   parse value LastSearchArgs with (delim)SearchString(delim)UserOptions
   UserOptions = strip( UserOptions, 'T', delim)

   -- Get current search direction
   Minuspos = lastpos( '-', UserOptions)
   Pluspos  = lastpos( '+', UserOptions)
   if Minuspos > Pluspos then
      fForward = 0
   else
      fForward = 1
   endif

   -- Always search in entire files, not in mark
   if pos( 'M', UserOptions) > 0 | pos( 'A', UserOptions) = 0 then
      -- Remove 'M' from UserOptions
      do forever
         pv = verify( upcase( UserOptions), 'M', 'M')
         if pv = 0 then
            leave
         endif
         UserOptions = delstr( UserOptions, pv, 1)
      enddo
      -- Append 'A'
      UserOptions = UserOptions'A'
      call SetLastSearchArgs( delim''SearchString''delim''UserOptions)
   endif

   -- Get LastSearchArgs from ini, remove 'T' and 'B' options
   'RepeatFind'

   if rc = 0 then
      -- Next occurrence found in current file
      return
   endif

   -- Not found in current file: search in other files
   getfileid fid
   startfid = fid
   do forever

      -- Next file
      if fForward = 1 then
         nextfile
      else
         prevfile
      endif
      getfileid fid
      activatefile fid

      call pSave_Pos( savedpos)
      -- Start from top of file
      if fForward = 1 then
         top
         .col = 1
      else
         bottom
         endline
      endif

      -- Get LastSearchArgs from ini, remove 'T' and 'B' options
      'RepeatFind'

      if rc = 0 then
         -- Found
         'HookAdd fileswitchedonce postme postme HighlightMatch'  -- additionally required to highlight after file switching
         if fid = startfid then
            'SayHint String only found in this file.'
         else
            sayerror 0  -- flush the message
         endif
         leave
      else
         -- Not found
         call pRestore_Pos( savedpos)
         if fid = startfid then
            'SayError String not found in any file of the ring.'
            leave
         else
            -- Search next file
         endif
      endif

   enddo
   activatefile fid

; ---------------------------------------------------------------------------
defc RepeatChangeAllFiles, RingChange
   universal stay

   LastChangeArgs = GetLastChangeArgs()
   delim = substr( LastChangeArgs, 1, 1)  -- get 1st delimiter
   parse value LastChangeArgs with (delim)SearchString(delim)ReplaceString(delim)UserOptions
   UserOptions = strip( UserOptions, 'T', delim)
   SavedOptions = UserOptions

   -- Get current search direction
   Minuspos = lastpos( '-', UserOptions)
   Pluspos  = lastpos( '+', UserOptions)
   if Minuspos > Pluspos then
      fForward = 0
   else
      fForward = 1
   endif

   -- Always search in entire files, not in mark
   if pos( 'M', UserOptions) > 0 | pos( 'A', UserOptions) = 0 then
      -- Remove 'M' from UserOptions
      do forever
         pv = verify( upcase( UserOptions), 'M', 'M')
         if pv = 0 then
            leave
         endif
         UserOptions = delstr( UserOptions, pv, 1)
      enddo
      -- Append 'A'
      UserOptions = UserOptions'A'
   endif

   -- Replace all occurrences, not only next
   if pos( '*', UserOptions) = 0 then
      -- Append '*'
      UserOptions = UserOptions'*'
   endif

   -- Write LastChangeArgs to ini
   if UserOptions <> SavedOptions then
      call SetLastChangeArgs( delim''SearchString''delim''ReplaceString''delim''UserOptions)
   endif

   getfileid fid
   startfid = fid
   ChangeCount = 0
   do forever

      call pSave_Pos( savedpos)
      -- Start from top of file
      if fForward = 1 then
         top
         .col = 1
      else
         bottom
         endline
      endif

      -- Get LastChangeArgs from ini, remove 'T' and 'B' options
      'RepeatChange'

      if rc = 0 then
         -- Found
         ChangeCount = ChangeCount + 1
         if stay then
            call pRestore_Pos( savedpos)
         endif
      elseif rc = -257 then  -- Invalid number of parameters
         return rc
      else
         -- Not found
         call pRestore_Pos( savedpos)
      endif

      -- Next file
      if fForward = 1 then
         nextfile
      else
         prevfile
      endif

      getfileid fid
      if fid = startfid then
         leave
      endif

   enddo

   if ChangeCount = 1 then
      files = 'file.'
   else
      files = 'files.'
   endif
   'SayHint String changed in' ChangeCount files

; ---------------------------------------------------------------------------
defc FindNextAllFiles
   'SearchDirection F'
   'RepeatFindAllFiles'

; ---------------------------------------------------------------------------
defc FindPrevAllFiles
   'SearchDirection R'
   'RepeatFindAllFiles'

; ---------------------------------------------------------------------------
defc FindMark
   if FileIsMarked() then
      -- Get active mark coordinates and fileid
      getmark FirstLine, LastLine, FirstCol, LastCol, MarkFileid
      if LastLine <> FirstLine then
         -- Take up to one line
         LastLine = FirstLine
         endline
         LastCol = .col
      endif
      SearchString = substr( textline( FirstLine ), FirstCol, LastCol - FirstCol + 1)
      if SearchString <> '' then
         'l '\1''SearchString
      endif
   else
      sayerror -280  -- Text not marked
   endif

; ---------------------------------------------------------------------------
; Find word under cursor -- if arg( 1) > 0: -- under pointer.
defc FindWord
   -- If arg( 1) specified and > 0: Set cursor to pos of pointer.
   if arg( 1) then
      'MH_GoToPosition'
   endif
   lrc = 1
   StartLine = .line
   StartCol  = .col
   call pEnd_Word()
   LastCol = .col
   call pBegin_Word()
   FirstCol = .col
   -- Start search after current word
   .col = LastCol + 1
   SearchString = substr( textline( StartLine), FirstCol, LastCol - FirstCol + 1)
   if SearchString <> '' then
      'l '\1''SearchString
      lrc = rc
   endif
   if lrc <> 0 then
      .col = StartCol
   endif

; ---------------------------------------------------------------------------
; Find identifier under cursor -- if arg( 1) > 0: -- under pointer.
defc FindToken
   -- If arg( 1) specified and > 0: Set cursor to pos of pointer.
   if arg( 1) then
      'MH_GoToPosition'
   endif
   lrc = 1
   call pSave_Pos( savedpos)
   if Find_Token( startcol, endcol) then
      -- find_token returns first and last col of the found string. Therefore
      -- search shall start from 1 col behind.
      .col = endcol + 1
      -- The standard cmd 'locate' won't set standard rc. The standard 'locate'
      -- always returns rc = ''.
      -- Standard commands that don't change rc (are there more?):
      --    'locate', 'edit'
      -- This is fixed in NEPMD.
      -- The standard commands 'quit', 'save', 'name',... do change rc.
      -- Therefore 'locate' now calls the proc locate, that sets rc correctly.
      -- rc is the rc from 'xcom locate'.
      -- (rc is a universal var, that doesn't need the universal definition.)
      'l '\1''substr( textline( .line), startcol, (endcol - startcol) + 1)
      lrc = rc
   endif
   --sayerror 'defc findword: lrc = 'lrc
   if lrc <> 0 then  -- if not found
      call pRestore_Pos( savedpos)
   endif

; ---------------------------------------------------------------------------
; Find a token around the cursor.
defproc Find_Token( var startcol, var endcol)
   if arg( 3) = '' then
      token_separators = ' ~`!%^&*()-+=][{}|\:;?/><,''"'\t
   else
      token_separators = arg( 3)
   endif
   if arg( 4) = '' then
      diads = '-> ++ -- << >> <= >= && || += -= *= /= %= = &= |= :: /* */'
   else
      diads = arg( 4)
   endif
   trailing_token_separators = '.:,;!?'
   getline line
   len = length( line)
   if .col > len | pos( substr( line, .col, 1), ' '\t) then
      -- Past end of line, or over whitespace
      return
   endif
   endcol = verify( line, token_separators, 'M', .col)
   if endcol = .col then
      -- On an operator
      startcol = endcol
      if wordpos( substr( line, startcol, 2), diads) then
         -- On first character
         endcol = endcol + 1
      elseif .col > 1 then
         if wordpos( substr( line, endcol-1, 2), diads) then
            -- On last character
            startcol = startcol - 1
         endif
      endif
      return 2
   endif
   if endcol then
      endcol = endcol - 1
   else
      endcol = len
   endif
   if pos( substr( line, endcol, 1), trailing_token_separators) then
      endcol = endcol - 1
   endif
   startcol = verify( reverse( line), token_separators, 'M', len - .col + 1)
   if startcol then
      startcol = len - startcol + 2
   else
      startcol = 1
   endif
   return 1

; ---------------------------------------------------------------------------
defc ShowSearch
   getsearch CurSearch

   Next = 'Last search = ['cursearch'], last search args = ['GetLastSearchArgs()']' ||
          ', last change args = ['GetLastChangeArgs()']'
   'SayHint' Next
   dprintf( Next)

; ---------------------------------------------------------------------------
defc SetScrollAfterLocate
   KeyPath = '\NEPMD\User\CursorPos\ScrollAfterLocate'
   DefaultValue = '26 -4'
   -- if executed with a num as arg
   parse value arg( 1) with topval botval
   if IsNum( topval) & IsNum( botval) then
      call WriteConfigKey( KeyPath, arg( 1))
      if isadefproc( 'MenuText_scrollafterlocate') then
         MenuText_scrollafterlocate()
      endif
      return
   endif
   -- else open entrybox
   Title   = 'Configure line position on screen after locate'
   Text    = 'Enter number of lines from top and from bottom (negative values for vspace).'
   IniValue = QueryConfigKey( KeyPath)
   If IniValue = 0 | IniValue = '' then
      DefaultButton = 3
   elseif IniValue < 0 then
      DefaultButton = 2
   else
      DefaultButton = 1
   endif
   -- strip + or -
   IniValue = strip( IniValue)
   parse value EntryBox( Title,
                         '/'OK__MSG'/~Reset/'CANCEL__MSG,  -- max. 4 buttons
                         IniValue,
                         '',
                         260,
                         atoi( DefaultButton)  ||
                         atoi( 0000)           ||  -- help id
                         GethWndC( APP_HANDLE) ||
                         Text) with Button 2 NewValue \0
   -- strip + or -
   NewValue = strip( NewValue)
   if Button = \1 then
      'SetScrollAfterLocate' NewValue
      return
   elseif Button = \2 then
      'SetScrollAfterLocate 'DefaultValue
      return
   elseif Button = \3 then
      return
   endif

; ---------------------------------------------------------------------------
defc GotoLineDlg
   Title = 'Go to line'
   Text  = 'Enter line number and optionally a column number:'
   --Text  = Text''copies( ' ', max( 100 - length(Text), 0))
   Entry = ''
   parse value EntryBox( Title,
                         '',
                         Entry,
                         0,
                         240,
                         atoi(1) || atoi(0) || atol(0) ||
                         Text) with button 2 NewLine \0
   NewLine = strip( NewLine)
   if button = \1 & NewLine <> '' then
      'GoTo' NewLine
   endif

; ---------------------------------------------------------------------------
const
compile if not defined( MATCH_MAX_LINES)
   MATCH_MAX_LINES = 200
compile endif
compile if not defined( MATCH_MAX_LOOPS)
   MATCH_MAX_LOOPS = 50
compile endif

defproc Match( OpenStr)
   universal CurKey
   universal PrevKey
   universal PrevMatchData

   StartLine = .line
   StartCol  = .col
   OpenLine = 0
   OpenCol  = 0
   lrc = 1

   CloseStr = arg( 2)
/*
   if CloseStr <> '' then
      call NextCmdAltersText()
      -- Type the char
      call Process_Keys( CloseStr)
   endif
*/

   -- Omit search for repeated keys after an unsuccessful search
   -- in other cases
   getline ThisLine
   RightLen = length( substr( ThisLine, .col))
   getfileid Fid
   parse value PrevMatchData with PrefFid PrevLine PrevRc PrevRightLen

   fSearch = 0
   if 0 then
   elseif RightLen = PrevRightLen & CurKey = PrevKey then
      fSearch = 0
   elseif CloseStr = '' then
      fSearch = 1
   elseif PrevRc = 0 then
      fSearch = 1
   elseif PrefFid PrevLine <> Fid StartLine then
      fSearch = 1
   elseif CurKey <> PrevKey then
      fSearch = 1
   endif

   if fSearch then
      display -3  -- turn off non-critical error messages and screen updates
                  -- Note: SayHint uses temp. 'display -8' to disable message box saving.
                  --       That results in 'display -11'.
      call pSave_Pos( ScreenPos)
      sayerror 0

      lrc = pAssist( 0, 0, MATCH_MAX_LINES, MATCH_MAX_LOOPS)

      PrevMatchData = Fid StartLine lrc RightLen
      if not lrc then
         AssistLine = textline( .line)
         OpenLine = .line
         OpenCol  = .col
      endif

      call pRestore_Pos( ScreenPos)
      display 3  -- turn on non-critical error messages and screen updates

      if lrc then
         -- Let pAssist do the error msg
      elseif OnScreen( OpenLine, OpenCol) then
         -- Highlight it
         EndCol = OpenCol + length( OpenStr) - 1
         'PostMe CircleIt' OpenLine OpenCol EndCol
      else
         -- Opening character not on screen, so tell user in message area where it is
         ReportLine = leftstr( AssistLine, OpenCol + length( OpenStr) - 1)
         if (length( ReportLine) > 20) then
            ReportLine = leftstr( ReportLine, 20) "..."
         endif
         'SayHint Line' OpenLine', column' OpenCol':' ReportLine
      endif
   endif

   return lrc

; ---------------------------------------------------------------------------
defc Match
   rc = Match( arg( 1))

; ---------------------------------------------------------------------------
; Syntax: rcx = SearchReplaceLine( SearchStr, ReplaceStr, ReplaceTimes
;                       [, SearchOpt [, Line , [StartCol [, EndCol]]]])
; Returns:
;    rcx = 0  line changed
;    rcx = 1  line not changed
; Default values for the optional args:
;    SearchOpt = ''
;    Line      = current line
;    StartCol  = current column
;    EndCol    = line end
; o  If '' is specified as SearchStr, the entire string from StartCol to
;    EndCol is replaced.
; o  Replaces a search mask on one line only.
; o  Found SearchStr is replaced ReplaceTimes times.
; o  If ReplaceTimes is empty or 0, all occurrences between StartCol and
;    EndCol are replaced.
; o  If StartCol is empty, the current .col is used.
; o  If EndCol is empty or -1, the rest of the line is searched.
; o  Any output is suppressed. Cursor and scroll pos. and previous search
;    args are restored after processing.
; o  This can be used instead of the replaceline statement to preserve
;    attributes (e.g. bookmarks).
defproc SearchReplaceLine( SearchStr, ReplaceStr, ReplaceTimes)
   if ReplaceTimes = '' then
      ReplaceTimes = 0
   endif
   SearchOpt = arg( 4)
   Line      = arg( 5)
   if Line = '' then
      Line = .line
   endif

   -- Ensure that Line exists
   LastLine = .last
   if Line > LastLine then
      do l = 1 to Line - LastLine
         insertline '', .last + 1
      enddo
   endif

   StartCol  = arg( 6)
   if StartCol = '' then
      StartCol = .col
   endif
   EndCol    = arg( 7)
   if EndCol = '' then
      EndCol = -1  -- -1 means: search until end of line
   endif
   rcx = 1

   fGrep = 0
   if pos( 'g', lowcase( SearchOpt)) | pos( 'x', lowcase( SearchOpt)) then
      fGrep = 1
   endif

   pSave_Pos( savepos)
   getsearch savesearch

   if SearchStr = '' then
      LineStr = textline( Line)
      if EndCol = -1 | EndCol >= 1599 then
         SearchStr = substr( LineStr, StartCol)
      else
         SearchStr = substr( LineStr, StartCol, Max( 0, EndCol - StartCol + 1))
      endif
   endif

   -- Handle empty SearchStr: Insert a '*' to make xcom l and xcom c work
   if SearchStr = '' then
      fInsertToggled = 0
      if insertstate() then
         inserttoggle
         fInsertToggled = 1
      endif

      .col   = StartCol
      .lineg = Line
      SearchStr = '*'
      TypeChars( SearchStr)

      if fInsertToggled then
         inserttoggle
      endif
   endif
   --dprintf( 'SearchReplaceLine: SearchStr = ['SearchStr']')

   .col   = StartCol
   .lineg = Line
   ProcessCount = 0
   do forever
      ProcessCount = ProcessCount + 1
      -- Check processing number
      if ReplaceTimes > 0 then
         if ProcessCount > ReplaceTimes then
            leave
         endif
      endif

      display -14
      -- Search
      'xcom l '\1''SearchStr''\1''SearchOpt
      display 14

      -- Check rc
      lrc = rc
      if lrc then
         leave
      endif
      -- Check pos.
      if .line <> Line then
         leave
      endif
      if EndCol > -1 then
         if .col > EndCol then
            leave
         endif
      endif

      -- Get line length before replace
      if fGrep then
         LineStr = textline( .line)
         LineLen = length( LineStr)
      endif

      display -14
      -- Replace
      'xcom c '\1''SearchStr''\1''ReplaceStr''\1''SearchOpt
      display 14

      -- Check rc
      crc = rc
      if crc then
         leave
      endif

      rcx = 0  -- line changed

      -- Move cursor after replace, adapt EndCol
      if fGrep then
         NewLineStr = textline( .line)
         NewLineLen = length( NewLineStr)
         DiffLen = NewLineLen - LineLen
      else
         DiffLen = length( ReplaceStr) - length( SearchStr)
      endif
      .col = .col + DiffLen
      if EndCol > -1 then
         EndCol = EndCol + DiffLen
      endif

      -- Check pos.
      if EndCol > -1 then
         if .col > EndCol then
            leave
         endif
      endif
   enddo

   setsearch savesearch
   pRestore_Pos( savepos)

   return rcx

; ---------------------------------------------------------------------------
; Syntax: pReplaceLine( ReplaceStr [, LineNum [, FileId]])
; o  This can be used instead of the replaceline statement to preserve
;    attributes (e.g. bookmarks). It has the same args.
defproc pReplaceLine( ReplaceStr)
   LineNum = arg( 2)
   if LineNum = '' then
      LineNum = .line
   endif

   getfileid StartFid
   fSwitchFile = 0
   ReplaceFid = arg( 3)
   if ReplaceFid = '' then
      -- nop
   elseif ReplaceFid <> StartFid then
      fSwitchFile = 1
   endif

   if fSwitchFile then
      activatefile ReplaceFid
   endif

   --dprintf( 'ReplaceStr = ['ReplaceStr'], LineNum = 'LineNum)
   rcx = SearchReplaceLine( '', ReplaceStr, 1, '', LineNum, 1)
   if fSwitchFile then
      activatefile StartFid
   endif
   return

